﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.Net.Sockets;
using System.Net;
using System.IO;
using System.Threading;

namespace AVREmulator
{
    public partial class Connections : Form
    {
        private TcpListener server;
        private TcpClient connection;
        private Thread connectionLoopThread;
        private NetworkStream stream;

        struct MessageType
        {
            public string Header;
            // returns the control to add to the log list
            public delegate Control MessageHandler(byte[] firstPacket, int packetLength);
            public MessageHandler Handler;

            public MessageType(string header, MessageHandler handler)
            {
                Header = header;
                Handler = handler;
            }
        }

        private MessageType[] messageTypes;

        public Connections()
        {
            InitializeComponent();

            messageTypes = new MessageType[] { 
                new MessageType( "PUT /screen", ScreenMessageHandler ),
                new MessageType( "PUT /font", null ),
                new MessageType( "PUT /color", null ),
                new MessageType( "POST ", null ),
                new MessageType( "GET ", null ),
            };
        }

        int SkipHeader(byte[] message)
        {
            string headerPart = Encoding.ASCII.GetString(message, 0, 256);
            return headerPart.IndexOf("\r\n\r\n");
        }

        Control ScreenMessageHandler(byte[] firstPacket, int packetLength)
        {
            byte[] totalScreenUpdate = new byte[ScreenBuffer.Size];

            int startIndex = SkipHeader(firstPacket);
            if (startIndex == -1)
            {
                return null;
            }

            bool commitBuffers = false;
            // check for the COMMIT flag
            if( Encoding.ASCII.GetString(firstPacket, 0, 256).IndexOf("swap_changed") != -1 ) {
                commitBuffers = true;
            }

            startIndex += 4;
            int remainingPacketSize = packetLength - startIndex;
            Array.Copy(firstPacket, startIndex, totalScreenUpdate, 0, remainingPacketSize);

            int remainingSize = ScreenBuffer.Size - remainingPacketSize;
            int bytesRead = stream.Read(totalScreenUpdate, remainingPacketSize, remainingSize);
            if (bytesRead != remainingSize)
            {
                return null;
            }
            return new ScreenMessageViewer(new ScreenBuffer(totalScreenUpdate), commitBuffers);
        }

        private Label CreateLabelEntry(string text)
        {
            Label label = new Label();

            label.AutoSize = true;
            label.BorderStyle = System.Windows.Forms.BorderStyle.None;
            label.BackColor = System.Drawing.SystemColors.ControlLightLight;
            label.Text = text;

            return label;
        }

        private delegate void AddEntryCallback(Control control);
        public void AddLogEntry(Control control)
        {

            // InvokeRequired required compares the thread ID of the
            // calling thread to the thread ID of the creating thread.
            // If these threads are different, it returns true.
            if (layout.InvokeRequired)
            {
                AddEntryCallback d = new AddEntryCallback(AddLogEntry);
                layout.Invoke(d, control);
            }
            else
            {
                control.Dock = System.Windows.Forms.DockStyle.Top;
                layout.Controls.Add(control);
            }
        }

        //public void CloseConnection()

        private void ConnectionLoop()
        {
            while (true)
            {
                AddLogEntry(CreateLabelEntry("Waiting for Connection on " + server.LocalEndpoint.ToString()));
                //server.BeginAcceptTcpClient(AsyncAcceptHandler, null);
                connection = server.AcceptTcpClient();

                AddLogEntry(CreateLabelEntry("Connection Accepted from " + connection.Client.RemoteEndPoint.ToString()));

                stream = connection.GetStream();

                // 1260: max packet size supported by the AVR atm
                byte[] buffer = new byte[1260];
                int bytesReceived;
                try
                {
                    while (connection.Connected && (bytesReceived = stream.Read(buffer, 0, buffer.Length)) != 0)
                    {
                        string text = Encoding.ASCII.GetString(buffer, 0, bytesReceived);
                        bool validMessage = false;
                        Control logEntry = null;
                        foreach (MessageType messageType in messageTypes)
                        {
                            if (text.StartsWith(messageType.Header))
                            {
                                validMessage = true;
                                if (messageType.Handler != null)
                                {
                                    logEntry = (Control) layout.Invoke( messageType.Handler, buffer, bytesReceived);
                                }
                                break;
                            }
                        }
                        if (!validMessage || logEntry == null)
                            logEntry = new LogRawMessage("Unknown Message Received (" + bytesReceived + " bytes)", text);
                        AddLogEntry(logEntry);
                    }
                }
                catch (IOException)
                {
                }

                AddLogEntry(CreateLabelEntry("Connection Closed (" + connection.Client.RemoteEndPoint.ToString() + ")"));
                stream.Close();
            }
        }

        private void Connections_Load(object sender, EventArgs e)
        {
            layout.Controls.Clear();

            IPEndPoint endPoint = new IPEndPoint(IPAddress.Loopback, 80);
            server = new TcpListener(endPoint);

            // Start listening for connection requests
            server.Start();

            connectionLoopThread = new Thread(ConnectionLoop);
            connectionLoopThread.Start();

            //new FPGAEmulator().Show();
        }

        private void Connections_FormClosed(object sender, FormClosedEventArgs e)
        {
            if (connection != null)
                connection.Close();

            connectionLoopThread.Abort();
            server.Stop();
        }
    }
}
