/*****************************************************************************
*  "A Very Simple Application" from the uIP 0.9 documentation
*****************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <avr/pgmspace.h>

#include "config.h"
#include "main.h"
#include "app.h"
#include "uip.h"
#include "fpga_serial.h"
#include "webclient.h"
#include "rawtransmission.h"
#include "packetbuffer.h"
#include "buffer.h"
#include "logservice.h"

typedef struct {
	const char *name;
	ServicePacketHandler handler;
} Service;

// register network services here
static Service services[] = {
	{ "Log Get", LogService_handleGetPacket },
	{ "WebClient Get", WebClient_handleGetPacket },
	{ "WebClient Post", WebClient_handlePostPacket },
	{ "RawTransmission Screen", RawTransmission_handleScreenPacket },
	{ "RawTransmission Font", RawTransmission_handleFontPacket },
	{ "RawTransmission Color", RawTransmission_handleColorPacket },
#ifdef BUFFER_RESET_PACKET
	{ "Buffer Reset", Buffer_handleResetPacket }
#endif
};

#define servicesCount (sizeof( services ) / sizeof( *services ))

// used for ACKs and RESENDs
static int16_t lastSendService[ UIP_CONNS ];
// used for NEWDATA, CLOSED
static int16_t morePacketsService[ UIP_CONNS ];

static int16_t currentService;

static bool server_processPacketBuffer(void);
//static const char * server_getStringForAnswer( ServiceAnswer answer );

// uses currentService!
static void server_printServiceMessage( const unsigned char *type, unsigned serviceIndex ) {
#ifdef SERVER_LOG
	if( serviceIndex < servicesCount ) {
		xprintf_P( PSTR( "--%s (%i) -- Service: %s\r\n" ), type, uip_getcurindex(), services[ serviceIndex ].name );
	}
	else {
		if( serviceIndex == -1 ) {
			xprintf_P( PSTR( "--%s (%i) -- No Service Assigned\r\n" ), type, uip_getcurindex() );
		} else {
			xprintf_P( PSTR( "--%s (%i)\r\n" ), type, uip_getcurindex() );
		}
	}
#endif
}

void Service_Send( const unsigned char *packet, uint16_t length ) {
	lastSendService[ uip_getcurindex() ] = currentService;
	uip_send( packet, length );

	server_printServiceMessage( "Send", currentService );
#ifdef ECHO_SENT
	HTTP_echoPacketHeader( packet, length );
#else
	HTTP_echoPacketInfo( packet, length );
#endif
}

// returns NULL on EOF
unsigned char * HTTP_skipHeader( const unsigned char *packet, uint16_t length ) {
	for( uint16_t i = 0 ; *packet && i < length ; packet++, i++ ) {
		if( *packet == '\r' ) {
			if( strncmp( packet, "\r\n\r\n", 4 ) == 0 ) {
				return packet + 4;
			}
		}
	}
	return NULL;
}

void HTTP_echoText( const unsigned char *data, uint16_t length ) {
	for( uint16_t i = 0 ; i < length && i < 512; i++ ) {
		if( !iscntrl( data[ i ] ) ) {
			xputc( data[ i ] );
		} else if( data[ i ] == '\r' ) {
			xputs( "\\r" );
		} else if( data[ i ] == '\n' ) {
			xputs( "\\n\r\n" );
		}
	}
}

void HTTP_echoPacket( const unsigned char *packet, uint16_t length ) {
	xprintf_P( PSTR( "@SoM@ (from %i: %i bytes total - excerpt)--" ), uip_getcurindex(), length );

	HTTP_echoText( packet, length );
	MSG( "--@EoM@" );
}

void HTTP_echoPacketHeader( const unsigned char *packet, uint16_t length ) {
	const unsigned char *data = HTTP_skipHeader( packet, length );

	if( data ) {
		uint16_t headerLength = data - packet;
	
		xprintf_P( PSTR( "@SoM@ (from %i: %i bytes total - %i bytes HTTP header)--" ), uip_getcurindex(), length, headerLength );

		HTTP_echoText( packet, headerLength );
		MSG( "--@EoM@" );
	}
	else {
		HTTP_echoPacket( packet, length );
	}
}

void HTTP_echoPacketInfo( const unsigned char *packet, uint16_t length ) {
	const unsigned char *data = HTTP_skipHeader( packet, length );
	
	if( data ) {
		uint16_t headerLength = data - packet;

		xprintf_P( PSTR( "HTTP Message from %i: %i bytes total - %i bytes HTTP header\r\n" ), uip_getcurindex(), length, headerLength );
	}
	else {
		xprintf_P( PSTR( "Raw Message from %i: %i bytes total\r\n" ), uip_getcurindex(), length );
	}
}

void HTTP_echoPacketData( const unsigned char *packet, uint16_t length ) {
	const unsigned char *data = HTTP_skipHeader( packet, length );

	if( data ) {
		uint16_t dataLength = length - (data - packet);
	
		xprintf_P( PSTR( "@SoD@ (from %i: %i bytes total - %i bytes HTTP packet)--" ), uip_getcurindex(), length, dataLength );
		HTTP_echoText( data, dataLength );
		MSG( "--@EoD@" );
	}
	else {
		HTTP_echoPacket( packet, length );
	}	
}

void server_init(void)
{
	LogService_init();
	WebClient_init();
	RawTransmission_init();
		  
	PacketBuffer_init();
	Buffer_Init();

	uip_listen(HTONS(80));
}

// returns TRUE if the packet has been processed
static bool server_interpretAnswer( ServiceAnswer answer ) {
	int16_t currentConn = uip_getcurindex();

	switch( answer ) {
		default:
		case SA_PASS_ON:				
			morePacketsService[ currentConn ] = -1;
			return FALSE;
		case SA_FINISH:
			uip_close();
			// the current connection handles the close event
			morePacketsService[ currentConn ] = currentService;
			return TRUE;
		case SA_ACCEPT:
			morePacketsService[ currentConn ] = -1;
			return TRUE;
		case SA_MORE_PACKETS:
			morePacketsService[ currentConn ] = currentService;
			return TRUE;
	}
}

/*
static const char * server_getStringForAnswer( ServiceAnswer answer ) {
	switch( answer ) {
		default:
			return "<unknown>";
		case SA_PASS_ON:				
			return "pass on";
		case SA_FINISH:
			return "finish";
		case SA_ACCEPT:
			return "accept";
		case SA_MORE_PACKETS:
			return "more packets";
	}
}*/

void server_app(void)
{
	int8_t currentConn = uip_getcurindex();

    if (uip_connected()) {
        lastSendService[ currentConn ] = -1;
		morePacketsService[ currentConn ] = -1;

		// init the packet buffer... !
		PacketBuffer_initConnection();
		
		server_printServiceMessage( "OPEN", -1 );
	}
	if( uip_acked() ) {
		// set for Service_Send and interpretAnswer
		currentService = lastSendService[ currentConn ];
		server_printServiceMessage( "ACK", currentService );

		if( currentService != -1 ) {
			// reset lastSendService now because Service_Send could be called later
			lastSendService[ currentConn ] = -1;

			ServiceAnswer answer = services[ currentService ].handler( SR_ACKED, NULL, 0 );
			// valid answers: FINISH or anything 
			if( answer == SA_FINISH ) {
				uip_close();
				server_printServiceMessage( "FINISH", currentService );
			}
			/* all other answers are ok */	
		}
	}
	if( uip_rexmit() ) {
		// set for Service_Send and interpretAnswer
		currentService = lastSendService[ currentConn ];
		server_printServiceMessage( "RESEND", currentService );

		if( currentService != -1 ) {
			// reset lastSendService
			lastSendService[ currentConn ] = -1;

			ServiceAnswer answer = services[ currentService ].handler( SR_RESEND, NULL, 0 );
			// valid answers: FINISH or anything 
			if( answer == SA_FINISH ) {
				uip_close();
				server_printServiceMessage( "FINISH", currentService );
			}
			/* all other answers are ok */
		}		
	}
	if( uip_closed() || uip_aborted() || uip_timedout() ) {
		// set for Service_Send and interpretAnswer
		currentService = morePacketsService[ currentConn ];
		server_printServiceMessage( uip_aborted() ? "ABORTED" : uip_timedout() ? "TIMEDOUT" :"CLOSED", currentService );

		if( currentService != -1 ) {
			// no answer required/all answers ignored
			services[ currentService ].handler( SR_CLOSED, NULL, 0 );

			// reset morePacketsService
			morePacketsService[ currentConn ] = -1;
			// dont send anything on close.. ever
			uip_send( NULL, 0 );
		}
	}
	
    if (uip_newdata() ) {
		server_printServiceMessage( "PACKET", -2 );

		// push it to the packet buffer
		PacketBuffer_pushData( uip_appdata, uip_datalen() );

		while( PacketBuffer_getLength() > 0 && server_processPacketBuffer() )
			;

		if( PacketBuffer_getLength() > MIN_RAW_PACKET_SIZE ) {
			server_printServiceMessage( "IGNORED", -1 );
			// eat the packet
			PacketBuffer_popData( PacketBuffer_getLength() );
		}
		// otherwise give it another chance next time around
		PacketBuffer_makePersistent();
	}

	if (uip_poll()) {
		// poll call
		currentService = morePacketsService[ currentConn ];
		if( currentService != -1 ) {
			services[ currentService ].handler( SR_POLL, NULL, 0 );
		}
	}
}

static bool server_processPacketBuffer(void) {
	int8_t currentConn = uip_getcurindex();

#ifdef SERVER_LOG
#	ifdef ECHO_RECEIVED
		HTTP_echoPacketHeader( PacketBuffer_getData(), PacketBuffer_getLength() );
#	else
		HTTP_echoPacketInfo( PacketBuffer_getData(), PacketBuffer_getLength() );
#	endif
#endif

	// set for Service_Send and interpretAnswer
	currentService = morePacketsService[ currentConn ];

	if( currentService != -1 ) {
		// give the morePacketsService a chance to process the packet
		serviceDataProcessedLength = PacketBuffer_getLength();
		ServiceAnswer answer = services[ currentService ].handler( SR_NEXT_PACKET, PacketBuffer_getData(), serviceDataProcessedLength );
			
		if( server_interpretAnswer( answer ) ) {
			PacketBuffer_popData(serviceDataProcessedLength);
			server_printServiceMessage( "CONTINUED", currentService );
			return TRUE;
		}
	}
	// probe our services
	for( currentService = 0 ; currentService < servicesCount ; currentService++ ) {
		// set for Service_Send and interpretAnswer
		serviceDataProcessedLength = PacketBuffer_getLength();
		ServiceAnswer answer = services[ currentService ].handler( SR_NEW_PACKET, PacketBuffer_getData(), serviceDataProcessedLength );

		if( server_interpretAnswer( answer ) ) {
			PacketBuffer_popData(serviceDataProcessedLength);
			server_printServiceMessage( "ASSIGNED", currentService );
			return TRUE;
		}
	}

	return FALSE;
}
