#include <ctype.h>

#include "config.h"
#include "buffer.h"

#ifdef BUFFER_RESET_PACKET
#	include <string.h>
#endif

typedef struct {
	bool textMode;

	uint16_t size, remaining;

#ifdef BUFFER_LOG
	const char *name;
#endif
} ManagedBuffer;

static ManagedBuffer buffers[ FD_COUNT ] = { { 0 } };
static FPGA_destination currentDestination = FD_SCREEN;

static bool acquired = FALSE;

static uint8_t changedBufferMask = 0;
static bool commitRequested = FALSE;

static void Buffer_CommitChangedBuffers(void);

// init the buffers
void Buffer_Init(void) {
	buffers[ FD_SCREEN ].size = SCREEN_SIZE;
	buffers[ FD_FONT ].size = FONT_SIZE;
	buffers[ FD_COLOR ].size = COLOR_SIZE;

#ifdef BUFFER_LOG
	buffers[ FD_SCREEN ].name = "screen";
	buffers[ FD_FONT ].name = "font";
	buffers[ FD_COLOR ].name = "color";
#endif

	FPGA_init();
	// send a reset signal to the FPGA
	FPGA_sendCommand( FC_RESET, FM_SCREEN | FM_FONT | FM_COLOR );
}

// TODO: assert that acquired is TRUE in Release and reset remaining bytes or early-out?!?

// release the current buffer
void Buffer_Release(void) {
	ManagedBuffer *buffer;

	acquired = FALSE;

	buffer = buffers + currentDestination;
	if( buffer->textMode ) {
		// did we sent anything at all?
		if( buffer->remaining == SCREEN_WIDTH * SCREEN_HEIGHT ) {
			// nothing needs to be done then
			return;
		}

		if( buffer->remaining > 0 ) {
	#ifdef BUFFER_LOG
			xprintf_P( PSTR( "Flushing remaining %i characters in text buffer\r\n" ), buffer->remaining );
	#endif	
			// flush the buffer (better than keeping it in limbo)
			FPGA_flushData( currentDestination, ((int32_t) buffer->remaining) * SCREEN_BPS);
		}
	}
	else {
		// did we sent anything at all?
		if( buffer->remaining == buffer->size ) {
			// nothing needs to be done then
			return;
		}
	
		if( buffer->remaining > 0 ) {
	#ifdef BUFFER_LOG
			xprintf_P( PSTR( "Flushing remaining %i bytes in %s buffer\r\n" ), buffer->remaining, buffer->name );
	#endif
			// flush the buffer (better than keeping it in limbo)
			FPGA_flushData( currentDestination, ((int32_t) buffer->remaining) << 3);
		}
	}

	// update the changed buffer mask
	changedBufferMask |= FD_TO_FM( currentDestination );

	// should we commit 
	if( commitRequested ) {
		Buffer_CommitChangedBuffers();
	}
}

static void Buffer_CommitChangedBuffers(void) {
#ifdef BUFFER_LOG
	xprintf_P( PSTR( "Committing changed buffers: " ) );
	for( int i = 0 ; i < FD_COUNT ; i++ ) {
		if( changedBufferMask & FD_TO_FM( i ) ) {
			xputs( buffers[i].name );
			xputc( ' ' );
		}
	}
	xputs("\r\n");
#endif
	
	FPGA_sendCommand( FC_COMMIT, changedBufferMask );
	xputc('.');

	changedBufferMask = 0;
	commitRequested = FALSE;
}

void Buffer_RequestCommit(void) {
#ifdef BUFFER_LOG
	xprintf_P( PSTR( "Buffer commit requested\r\n" ) );
#endif

	if( acquired ) {
		commitRequested = TRUE;
	} else {
		Buffer_CommitChangedBuffers();
	}
}

// write data to the current buffer (assert that a writing is allowed..)
void Buffer_Write( const unsigned char *data, uint16_t length ) {
	// assert: we are not in textmode!
	ManagedBuffer *buffer = buffers + currentDestination;
	
	if( length == 0 || buffer->remaining == 0 ) {
		MSG( "--ERR: empty data or buffer done" );
		return;
	}
	
	// boundary check
	if( length > buffer->remaining ) {
		// sanatize
		length = buffer->remaining;
		
		MSG( "--ERR: Buffer overflow detected and avoided" );
	}
	
	FPGA_sendData( currentDestination, data, length );
	buffer->remaining -= length;
	
#ifdef BUFFER_LOG
	xprintf_P( PSTR( "Uploaded %i bytes to %s buffer - remaining: %i\r\n" ), length, buffer->name, buffer->remaining );
#endif
}

static void Buffer_AcquireInternal( FPGA_destination destination ) {
	currentDestination = destination;

	acquired = TRUE;

	// send a reset signal to the FPGA
	FPGA_sendCommand( FC_RESET, FM_SCREEN | FM_FONT | FM_COLOR );
}

void Buffer_Acquire( FPGA_destination destination ) {
	Buffer_AcquireInternal( destination );
	ManagedBuffer *buffer = buffers + currentDestination;
	
	buffer->remaining = buffer->size;
	buffer->textMode = FALSE;
}

void Buffer_AcquireTextMode() {
	Buffer_AcquireInternal( FD_SCREEN );
		
	// TODO: add a static global buffer >_>
	ManagedBuffer *buffer = buffers + currentDestination;
	buffer->remaining = SCREEN_WIDTH * SCREEN_HEIGHT;
	buffer->textMode = TRUE;
}

uint16_t Buffer_GetRemaining(void) {
	return (buffers[ currentDestination ].remaining);
}

//uint16_t Buffer_GetSize(void) {
//	return (buffers[ currentDestination ].size);
//}

bool Buffer_Done(void) {
	return (Buffer_GetRemaining() == 0);
}

bool Buffer_Acquired(void) {
	return acquired;
}

// advanced function to write encoded data to the screen
void Buffer_WriteEncodedScreenData( unsigned char frontColor, unsigned char backColor, unsigned char *data, uint16_t length ) {
	const unsigned char *in = data; 
	
	int16_t row, column;
	// assert: remaining > 0
	// remaining rows and columns in range (0,0)-(79,39)
	row = (buffers[ FD_SCREEN ].remaining - 1) / SCREEN_WIDTH;
	column = (buffers[ FD_SCREEN ].remaining - 1) % SCREEN_WIDTH;

	FPGA_beginData( FD_SCREEN );
	FPGA_setColor( frontColor, backColor /* + 16 */);
	for( uint16_t i = 0 ; i < length ; i++, in++ ) {
		if( column == -1 ) {
			column = SCREEN_WIDTH - 1;
			row -= 1;
			if( row == -1 ) {
				// we're done..
				break;
			}
		}

		if( iscntrl( *in ) ) {
			if( *in == '\0' ) {
				for( ; row >= 0 ; row-- ) {
					for( ; column >= 0 ; column-- ) {
						FPGA_pushChar( ' ' );
					}
					column = SCREEN_WIDTH - 1;
				}
				break;
			}
			else if( *in == '\n' ) {
				// fill remaining line
				for( ; column >= 0 ; column-- ) {
					FPGA_pushChar	( ' ' );
				}
			} else if( *in == '\t' ) {
				// tab (insert 4 spaces)
				for( uint16_t k = 4 ; k > 0 && column >= 0 ; k--, column-- ) {
					FPGA_pushChar( ' ' );
				}
			}
		} else {
			FPGA_pushChar( *in );
			column--;
		}
	}
	FPGA_endData();

	// regenerate remaining bytes
	buffers[ FD_SCREEN ].remaining = (1 + row * SCREEN_WIDTH + column);
#ifdef BUFFER_LOG
	xprintf_P( PSTR( "Characters remaining in screen buffer: %i\r\n" ), buffers[ FD_SCREEN ].remaining );
#endif
}

#ifdef BUFFER_RESET_PACKET
ServiceAnswer Buffer_handleResetPacket( ServiceRequest request, unsigned char *packet, uint16_t length ) {
	if( request == SR_NEW_PACKET ) {
		if( strncmp( packet, "PUT /reset", sizeof( "PUT /reset" ) - 1 ) == 0 ) {
		// there can be only one atm
			if( Buffer_Acquired() ) {
				Buffer_Release();
			}
			return SA_ACCEPT;
		}
	}
	return SA_PASS_ON;
}
#endif
