#include <stdlib.h>
#include <avr/pgmspace.h>
#include <stdint.h>
#include <stdio.h>
#include <string.h>

#include "config.h"
// TODO: move the destination stuff to another enum
#include "fpga_serial.h"
#include "buffer.h"
#include "rawtransmission.h"

#define PROTOCOL_SCREEN_STR "PUT /screen"
#define PROTOCOL_FONT_STR "PUT /font"
#define PROTOCOL_COLOR_STR "PUT /color"

#define PROTOCOL_SWAP_FLAG "swap_changed"

#define REPLACE( x ) 	(x)
#define STR( d )		REPLACE( d##_STR )

static unsigned char *packetPrefixes[ FD_COUNT ];

void RawTransmission_init(void) {
	packetPrefixes[ FD_SCREEN ] = PROTOCOL_SCREEN_STR;
	packetPrefixes[ FD_FONT ] = PROTOCOL_FONT_STR;
	packetPrefixes[ FD_COLOR ] = PROTOCOL_COLOR_STR;
}

// returns FALSE, when we're done
static bool RawTransmission_transmit( const unsigned char *data, uint16_t length ) {
	// assert: Buffer_Acquired() == TRUE
	int overflow = length - Buffer_GetRemaining();
	if( overflow > 0 ) {
		length -= overflow;
		serviceDataProcessedLength -= overflow;
		xputc('/');
	}

	Buffer_Write( data, length );

	if( Buffer_Done() ) {
		Buffer_Release();
		return FALSE;
	}

	return TRUE;
}

static bool RawTransmission_checkForSwapFlag( const unsigned char *data, uint16_t length ) {
	// skip whitespace
	for( ; (*data == ' ' || *data == '\t') && length > 0 ; data++, length-- )
		;

	// check for the flag
	
	//   avoid buffer overruns
	if( length < sizeof( PROTOCOL_SWAP_FLAG ) - 1 ) {
		return FALSE;
	}

	if( strncmp( data, PROTOCOL_SWAP_FLAG, sizeof( PROTOCOL_SWAP_FLAG ) - 1 ) == 0 ) {
		return TRUE;
	}
	return FALSE;
}

ServiceAnswer RawTransmission_handleGenericPacket( FPGA_destination destination, ServiceRequest request, unsigned char *packet, uint16_t length ) {
	if( request == SR_NEW_PACKET ) {
		uint16_t packetPrefixLength = strlen( packetPrefixes[ destination ] );

		if( strncmp( packet, packetPrefixes[ destination ], packetPrefixLength ) == 0 ) {
			// there can be only one atm
			if( !Buffer_Acquired() ) {
				unsigned char *data = HTTP_skipHeader( packet, length );
				if( !data ) {
					return SA_ACCEPT;
				}

				Buffer_Acquire( destination );

				// check for the swap flag (after acquiring the buffer to delay it till after the release of the buffer)
				if( RawTransmission_checkForSwapFlag( packet + packetPrefixLength, length - packetPrefixLength ) ) {
					Buffer_RequestCommit();
				}

				if( RawTransmission_transmit( data, length - (data - packet) ) ) {
					return SA_MORE_PACKETS;
				}
				else {
					return SA_ACCEPT;
				}
			}
			// eat the packet
			return SA_ACCEPT;				
		}
	}
	else if( request == SR_NEXT_PACKET ) {
		if( RawTransmission_transmit( packet, length ) ) {
			return SA_MORE_PACKETS;
		}
		else {
			return SA_ACCEPT;
		}
	}
	else if( request == SR_CLOSED ) {
		// TODO: code review.. is it possible that SR_CLOSED happens after everything is sent and a new transmission has begun? (concurrency?)
		Buffer_Release();
		return SA_OK;
	}
	return SA_PASS_ON;
}

ServiceAnswer RawTransmission_handleScreenPacket( ServiceRequest request, unsigned char *packet, uint16_t length ) {
	return RawTransmission_handleGenericPacket( FD_SCREEN, request, packet, length );
}

ServiceAnswer RawTransmission_handleFontPacket( ServiceRequest request, unsigned char *packet, uint16_t length ) {
	return RawTransmission_handleGenericPacket( FD_FONT, request, packet, length );
}

ServiceAnswer RawTransmission_handleColorPacket( ServiceRequest request, unsigned char *packet, uint16_t length ) {
	return RawTransmission_handleGenericPacket( FD_COLOR, request, packet, length );
}
