/*--------------------------------------------------------------------
 serial.c
 Functions for serial communication on the Reelbox
 (c) 2004 BayCom GmbH
 $Id: serial.c,v 1.3 2005/03/06 19:08:19 acher Exp $
 *-------------------------------------------------------------------*/

#include <avr/io.h>
#include <util/delay.h>
#include <avr/pgmspace.h>
#include <avr/interrupt.h>

#include <stdarg.h>
#include <stdio.h>

#include "config.h"
#include "serial.h"

#define RX_BUF_SIZE (1<<3)  // Power of 2
#define TX_BUF_SIZE (1<<9) // Power of 2
uint8_t rx_buf[RX_BUF_SIZE];
uint16_t rx_rdptr;
uint16_t rx_wrptr;

uint8_t tx_buf[TX_BUF_SIZE];
uint16_t tx_rdptr;
uint16_t tx_wrptr;

static uint16_t x_baud_delay_us;

/*------------------------------------------------------------------*/
void usart_init(uint16_t baud)
{
	x_baud_delay_us = (uint32_t)(10000 / baud * 100);
    UBRR1H = (unsigned char)(baud>>8);
    UBRR1L = (unsigned char)baud;
    UCSR1B = (1<<RXCIE)|(1<<TXCIE)|(1<<RXEN)|(1<<TXEN); /* Enable receiver and transmitter */
    UCSR1C = (3<<UCSZ0); /* 8N1 */
    rx_rdptr=rx_wrptr=0;
    tx_rdptr=tx_wrptr=0;
}
/*------------------------------------------------------------------*/
/* Receive interrupt */
SIGNAL(SIG_UART1_RECV)
{
    uint8_t a=UDR1;
    rx_buf[rx_wrptr]=a;
    rx_wrptr=(rx_wrptr+1)&(RX_BUF_SIZE-1);
}
/*------------------------------------------------------------------*/
uint16_t rx_get(void)
{
    uint16_t ret;
    cli();
    if (rx_rdptr==rx_wrptr)
    {
        ret=65535;
    }
    else
    {
        ret=rx_buf[rx_rdptr];
        rx_rdptr=(rx_rdptr+1)&(RX_BUF_SIZE-1);
    }
    sei();
    return ret;
}
/*------------------------------------------------------------------*/
/* Transmit interrupt */
SIGNAL(SIG_UART1_TRANS)
{
    if (tx_rdptr!=tx_wrptr)
    {
        UDR1=tx_buf[tx_rdptr];
        tx_rdptr=(tx_rdptr+1)&(TX_BUF_SIZE-1);
    }
}
/*------------------------------------------------------------------*/
void xputc(uint8_t c)
{
    cli();
    tx_buf[tx_wrptr]=c;
    tx_wrptr=(tx_wrptr+1)&(TX_BUF_SIZE-1);
    if ( UCSR1A & (1<<UDRE))
    {
        UDR1=tx_buf[tx_rdptr];
        tx_rdptr=(tx_rdptr+1)&(TX_BUF_SIZE-1);
    }
    sei();

	// avoid overwriting data that has not been sent yet
	//while( (tx_wrptr+1)&(TX_BUF_SIZE-1) == tx_rdptr )
	 {
		//_delay_us( x_baud_delay_us );
	}
	#ifdef ENABLE_LOGSERVICE
	{
		char tmp[2] = { c, 0 };
		LogService_add( tmp, 1 );
	}
	#endif
}
/*------------------------------------------------------------------*/
void xputs(char *s)
{
   while (*s)
        xputc(*s++);
}
/*------------------------------------------------------------------*/
void printhex_sub(uint8_t a)
{
    a&=15;
    if (a<10)
        xputc('0'+a);
    else
        xputc('a'+(a-10));
}
/*------------------------------------------------------------------*/
void printhex(uint16_t a)
{
    printhex_sub(a>>12);
    printhex_sub(a>>8);
    printhex_sub(a>>4);
    printhex_sub(a);
}
/*------------------------------------------------------------------*/
void printhex8(uint8_t a)
{
    printhex_sub(a>>4);
    printhex_sub(a);
}
/*------------------------------------------------------------------*/
void printhex_l(uint32_t a)
{
    printhex((uint16_t)(a>>16));
    printhex((uint16_t)(a));
}
/*------------------------------------------------------------------*/
#if defined( ENABLE_XPRINTFS ) || defined( ENABLE_LOGSERVICE )
void xprintf( const char *fmt, ... ) {
	extern unsigned char *pTextBuffer;
	int length;

	va_list params;

	va_start( params, fmt );
	length = vsnprintf( pTextBuffer, 256, fmt, params );
	va_end( params );

#ifdef ENABLE_XPRINTFS
	xputs( pTextBuffer );
#endif
#ifdef ENABLE_LOGSERVICE
	LogService_add( pTextBuffer, length );
#endif
}

void xprintf_P( const char *fmt, ... ) {
	int length;
	extern unsigned char *pTextBuffer;

	va_list params;

	va_start( params, fmt );
	length = vsnprintf_P( pTextBuffer, 256, fmt, params );
	va_end( params );

#ifdef ENABLE_XPRINTFS
	xputs( pTextBuffer );
#endif
#ifdef ENABLE_LOGSERVICE
	LogService_add( pTextBuffer, length );
#endif
}
#endif
