#include <stdlib.h>
#include <avr/pgmspace.h>
#include <stdint.h>
#include <stdio.h>
#include <string.h>

#include "config.h"
// TODO: move the destination stuff to another enum
#include "fpga_serial.h"
#include "buffer.h"
#include "webclient.h"

#define FLASH_DATA( dataIdentifier ) \
	static const unsigned char * dataIdentifier; \
	static const unsigned char flash_##dataIdentifier[] PROGMEM =

#define INIT_FLASH_DATA( dataIdentifier ) do {\
			(dataIdentifier) = malloc( sizeof( flash_##dataIdentifier ) );\
			memcpy_P( (unsigned char*) dataIdentifier, flash_##dataIdentifier, sizeof( flash_##dataIdentifier ) );\
			} while (0)

#define FLASH_DATA_LENGTH( dataIdentifier ) (sizeof( flash_##dataIdentifier ))

FLASH_DATA( webClientData ) {
#include "webclient.prefix.data"
,
#include "webclient.data"
};

FLASH_DATA( colorData ) {
// data is a byte
#define CHANNEL(data, position) (( ((unsigned char)(data) >> 4) & 15 ) << position)
#define COLORBYTE(high, low)	(CHANNEL( (high), 4) | CHANNEL( (low), 0) )

#define COLOR_TABLE_BEGIN()
#define COLOR_TABLE_END()
#define COLOR_TABLE_ENTRY( index, red, green, blue, blinking ) \
		COLORBYTE( red, green),\
		COLORBYTE( blue, (red) >> (blinking)),\
		COLORBYTE( (green) >> (blinking), (blue) >> (blinking) ),

#include "..\colortable.h"

#undef COLOR_TABLE_ENTRY
#undef COLOR_TABLE_END
#undef COLOR_TABLE_BEGIN

#undef POSITION_SHRUNK_DATA
};

FLASH_DATA( fontData ) {
#include "webclientfont.data"
};

#define webClientData_length	FLASH_DATA_LENGTH( webClientData )
#define colorData_length		FLASH_DATA_LENGTH( colorData )
#define fontData_length			FLASH_DATA_LENGTH( fontData )


static unsigned char urlBufferedChar = 0;
static uint8_t urlRemainingChars = 0;
static unsigned char frontColor = 0, backColor = 0;

static uint16_t sentBytes[ UIP_CONNS ];
static uint16_t ackedBytes[ UIP_CONNS ];

void WebClient_init(void) {
	INIT_FLASH_DATA( webClientData );
	INIT_FLASH_DATA( colorData );
	INIT_FLASH_DATA( fontData );
}


static inline uint8_t HexCharToInt( unsigned char c ) {
	if( '0' <= c && c <= '9' ) {
		return c - '0';
	}
	else if( 'a' <= c && c <= 'f' ) {
		return c - 'a' + 10;
	}
	else if( 'A' <= c && c <= 'F' ) {
		return c - 'A' + 10; 
	} else {
		return 0;
	}
}

// tested tuesday, 1 pm - DO NOT change anymore if possible!
// returns the length of the decoded data
uint16_t WebClient_decodeData( unsigned char *data, uint16_t length ) {
	unsigned char *in = data, *out = data;
	uint16_t i = length;
	
	if( urlRemainingChars ) {
		for( ; urlRemainingChars > 0 ; urlRemainingChars--, i-- ) {
			urlBufferedChar = (urlBufferedChar << 4) | HexCharToInt( *out++ );
		}
		*in++ = urlBufferedChar;
	}
	
	for( ; i > 0 ; ) {
		// & marks the beginning of the &eof= part and we're done when we see this
		if( *out == '&' ) {
			// done
			*in++ = 0;
			i--;
			break;			
		}
		else if( *out == '%' ) {
			// decode %HL as usual
			if( i >= 3 ){
				*in++ = (HexCharToInt( out[1] ) << 4) + HexCharToInt( out[2] );
				out += 3;
				i -= 3;
			}
			// if only two chars or less are left we have to remember the % or the % and the high byte
			else  {
				urlRemainingChars = 3 - i;
				// remember the high byte if two are left
				if( i == 2 ) {
					urlBufferedChar = HexCharToInt( out[ 1 ] ) << 4;
				}
				break;
			}
		}
		// normal character or + for space
		else {
			if( *out != '+' ) {
				*in++ = *out++;
			}
			else {
				*in++ = ' ';
				out++;
			}
			i--;
		}
	}
	
	return in - data;
}

static inline uint16_t min16( uint16_t a, uint16_t b ) {
	if( a < b ) 
		return a;
	return b;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////
// Service to send the webclient (as compressed html file) to a client (using multiple packets if necessary)

// returns FALSE if all has been sent
static ServiceAnswer WebClient_sendPacket(void) {
	uint8_t currentIndex = uip_getcurindex();

	ackedBytes[ uip_getcurindex() ] = sentBytes[ uip_getcurindex() ];
	sentBytes[ currentIndex ] = min16( ackedBytes[ currentIndex ] + uip_mss(), webClientData_length );
	
	uint16_t length = sentBytes[ currentIndex ] - ackedBytes[ currentIndex ];

//	sprintf( pTextBuffer, "mss: %i acked: %i sent: %i length: %i\r\n", uip_mss(), ackedBytes[ currentIndex ], sentBytes[ currentIndex ], length );
//	xputs( pTextBuffer );

	if( length > 0 ) {
		Service_Send( webClientData + ackedBytes[ currentIndex ], length );
		return TRUE;
	}
	return FALSE;
}

static void WebClient_startSending(void) {
	ackedBytes[ uip_getcurindex() ] = 0;
	sentBytes[ uip_getcurindex() ] = 0;

	WebClient_sendPacket();
}

static void WebClient_resendPacket(void) {
	uint8_t currentIndex = uip_getcurindex();
	Service_Send( webClientData + ackedBytes[ currentIndex ], sentBytes[ currentIndex ] - ackedBytes[ currentIndex ] );
}

ServiceAnswer WebClient_handleGetPacket( ServiceRequest request, unsigned char *data, uint16_t length ) {
	if( request == SR_NEW_PACKET ) {
		if( strncmp( data, WEBCLIENT_GET_CMD, sizeof( WEBCLIENT_GET_CMD ) - 1 ) == 0 ) {
			WebClient_startSending();
			return SA_ACCEPT;
		}
	}
	else if( request == SR_ACKED ) {
		if( WebClient_sendPacket() ) {
			return SA_OK;
		} else {
			return SA_FINISH;
		}
	}
	else if( request == SR_RESEND ) {
		WebClient_resendPacket();
		return SA_OK;
	}
	
	return SA_PASS_ON;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////
// Service to post data to screen (from the webclient)

// XX and YY are fill-up characters for the numbers
#define FRONT_COLOR_START	(sizeof( "text_color=" ) - 1)
#define BACK_COLOR_START	(FRONT_COLOR_START + sizeof( "XX&background_color=" ) - 1)
#define TEXT_START			(BACK_COLOR_START + sizeof( "YY&text=" ) - 1)
#define PARAMETER_MIN_LENGTH TEXT_START

#define fromDigit( c ) ( (c) - '0' )

// returns a pointer to the first char after ...data=
static unsigned char* WebClient_parseParameters( unsigned char *params, uint16_t length ) {
	frontColor = fromDigit( params[ FRONT_COLOR_START ] ) * 10 + fromDigit( params[ FRONT_COLOR_START + 1 ] );
	backColor = fromDigit( params[ BACK_COLOR_START ] ) * 10 + fromDigit( params[ BACK_COLOR_START + 1 ] );

	return params + TEXT_START;
}

// return the correct service answer
static ServiceAnswer WebClient_transmit( unsigned char *data, uint16_t length ) {
	// decode the encoded stream into same buffer
	uint16_t decodedLength = WebClient_decodeData( data, length );

	Buffer_WriteEncodedScreenData( frontColor, backColor, data, decodedLength );

	if( Buffer_Done() ) {
		Buffer_Release();

		// transmit the default color and font tables
		Buffer_Acquire( FD_COLOR );
		Buffer_Write( colorData, colorData_length );
		Buffer_Release();

		Buffer_Acquire( FD_FONT );
		Buffer_Write( fontData, fontData_length );
		Buffer_Release();

		// request the commit on the FPGA
		Buffer_RequestCommit();

		// send a reply to the webclient (to make it reload the page)
		Service_Send( WEBCLIENT_POST_REPLY, sizeof( WEBCLIENT_POST_REPLY ) - 1 );
	
		return SA_ACCEPT;
	}
	else {
		return SA_MORE_PACKETS;
	}
}

ServiceAnswer WebClient_handlePostPacket( ServiceRequest request, unsigned char *packet, uint16_t length ) {
	if( request == SR_NEW_PACKET ) {
		if( strncmp( packet, WEBCLIENT_SCREEN_CMD, sizeof( WEBCLIENT_SCREEN_CMD ) - 1 ) != 0 ) {
			return SA_PASS_ON;
		}

		if( Buffer_Acquired() ) {
			// eat the packet anyway
			return SA_ACCEPT;
		}

		// get the screen buffer (using the special text mode)
		Buffer_AcquireTextMode();

		// reset the url decoder
		urlRemainingChars = 0;
	
		// jump to the start over the header
		unsigned char *parameters = HTTP_skipHeader( packet, length );
		uint16_t parameterLength = length - (parameters - packet);

		// if the packet header is too big, ignore the packet, because it doesnt make sense to extend the state machine for that case
		if( !parameters || parameterLength <= TEXT_START ) {
			// eat the packet anyway
			return SA_ACCEPT;
		}

		// parse the parameters
		unsigned char *data = WebClient_parseParameters( parameters, parameterLength );
		uint16_t dataLength = length - (data - packet);
		return WebClient_transmit( data, dataLength );
	}
	else if( request == SR_NEXT_PACKET ) {
		return WebClient_transmit( packet, length );	
	}
	else if( request == SR_CLOSED ) {
		Buffer_Release();
		return SA_OK;
	}
	else if( request == SR_ACKED ) {
		return SA_FINISH;
	}
	else if( request == SR_RESEND ) {
		// send a reply to the webclient (to make it reload the page)
		Service_Send( WEBCLIENT_POST_REPLY, sizeof( WEBCLIENT_POST_REPLY ) - 1 );
		return SA_OK;
	}

	return SA_PASS_ON;
}
